#  SSheetCell.py --- Class encapsulating a spreadsheet cell
#
#  last modified: 01/22/04
#
#                     =====================================
#                                 Marc Provost
#                          (marc.provost@mail.mcgill.ca)
#
#
#                          McGill University, Montreal
#                     =====================================
#
#******************************************************************************

from formulaparser.Parser import *
from formulaparser.syntaxtrees.ASTVisitor import ASTVisitor
from formulaparser.syntaxtrees.AST import *

from SSheetData import SSheetData
from SSheetErrors import *
from Functions import *

#******************************************************************************
### class DependVisitor
class DependVisitor(ASTVisitor):
    """Class visiting an AST to create a cell dependency list
    """

    #INPUT: <None>
    #Order: O(1)
    #Info: This class will go through the AST and
    #      add all the references to other cells
    #      in dependsOn, a dictionnary mapping
    #      the tuple (col,row)->CellRef object
    def __init__(self):
        self.dependsOn={}

    def visitNumber(self, number):
        pass

    def visitFunction(self, function):
        pass

    def visitCellRef(self, cellRef):
        self.dependsOn[(cellRef.getColumn(),
                        cellRef.getRow())]=cellRef

    def visitRangeRef(self, rangeRef):
        for cell in rangeRef.getCellRefSet():
            self.dependsOn[(cell.getColumn(),
                            cell.getRow())]=cell

#******************************************************************************
### class EvalVisitor
class EvalVisitor(ASTVisitor):
    """Visit the formula AST and evaluate it
    """

    #INPUT: SSheetData
    #Order: O(1)
    #Info: This class will go through the AST and
    #      evaluate it.  A reference to SSheetData
    #      is needed to resolve and evaluate references to
    #      other cells.  If a reference does not exist in
    #      SSheetData, it evaluate to 0.
    #      Note that it is assumed that cycles were handled
    #      beforehand (i.e. This visitor is not executed on
    #      cycles).

    def __init__(self,sheet):
        self.stack=[]
        self.sheet=sheet
        self.valueError=0

    def visitNumber(self, number):
        self.stack.append(number.getValue())

    #Info: Execute the function on its arguments.
    #      Functions are implemented in the file
    #      Functions.py.  If a function is not
    #      found in the file, a FunctionNameError
    #      is raised.

    def visitFunction(self, function):
        tmp=self.stack[-len(function.getArgs()):]
        self.stack=self.stack[0:-len(function.getArgs())]

        args=[]
        if len(tmp)>1:
            import types
            for val in tmp:
                if type(val)==types.StringType:
                    self.valueError=1
                else:
                    args.append(val)
        else:
          args=tmp
        execStr="answer = "+function.getName()+"("+str(args)+")"
        try:
            exec execStr
        except NameError, n:
            fName=split(n[0],"'")
            raise FunctionNameError(fName[1])
        self.stack.append(answer)

    #Info: Evaluate the cell referenced by cellRef.
    #      If the cell does not exists in the SSheetData,
    #      consider its value to be 0

    def visitCellRef(self, cellRef):
        cell=self.sheet.getCell(cellRef.getColumn(),cellRef.getRow())
        if cell==None:
            self.stack.append(0)
        else:
            self.stack.append(cell.evaluate())

    #Info: Evaluate the cells referenced by rangeRef.
    #      If a cell does not exists in the SSheetData,
    #      consider its value to be 0

    def visitRangeRef(self, rangeRef):
        valueSet=[]
        for cellRef in rangeRef.getCellRefSet():
            cell=self.sheet.getCell(cellRef.getColumn(),
                                    cellRef.getRow())
            if cell==None:
                valueSet.append(0)
            else:
                valueSet.append(cell.evaluate())
        self.stack.append(valueSet)


class SSheetCell:
    """Class encapsulating a cell in the spreadsheet
    """

    #INPUT: <int> <int> <string>
    #Order: O(len(formulaStr))
    #Info: Given a column, row, and a fornula,
    #      Initialize the cell. TypeErrors are
    #      raised if the types of the arguments are
    #      incorrect.  The formula string is parsed,
    #      and an AST is created if there is no parse errors.
    #      Otherwise, a FormulaParseError is raised.  A dependency list
    #      (of other cells) is also generated using the DependVisitor
    #      on the AST.

    def __init__(self, col, row, formulaStr, data):
        if type(col)!=types.IntType:
            raise TypeError, "constructor expected an integer for the column"
        if type(row)!=types.IntType:
            raise TypeError, "constructor expected an integer for the row"
        if type(formulaStr)!=types.StringType:
            raise TypeError, "constructor expected a string for the formula"

        self.__col=col
        self.__row=row
        self.__SSheetData=data
        self.__value = 0
        self.__formulaStr=formulaStr

        #parsing the formula string, if necessary
        if formulaStr[0]=='=':
            astTuple=getAST(formulaStr+";")
            if not astTuple[0]:
	      # HV: pass the error msg string from the parser
              raise FormulaParseError(astTuple[2])
            self.__ast=astTuple[1]
            #creating the dependence list
            v=DependVisitor()
            self.__ast.accept(v)
            self.__dependsOn=v.dependsOn
        else:
            self.__ast=None
            self.__value=formulaStr
            self.__dependsOn={}

    #INPUT: <None>
    #OUTPUT: ASTNode
    #Order: O(1)
    #Info: Return the AST associated with this cell
    def getAST(self):
        return self.__ast

    #INPUT: <ASTNode>
    #OUTPUT: <None>
    #Order: O(1)
    #Info: Set the AST of this cell.  Will be used
    #      when cells are copied into other cells to
    #      avoid reparsing the same formula string.
    def setAST(self, ast):
        if(not isinstance(ast, ASTNode)):
            raise TypeError,"setAST was expecting ASTNode instance"
        self.__ast=ast

        #creating the dependence list
        v=DependVisitor()
        self.__ast.accept(v)
        self.__dependsOn=v.dependsOn

        self.__SSheetData.evaluate()

    #INPUT: <string>
    #OUTPUT: <None>
    #Order: O(len(formulaStr))
    #Info: Set the formulaStr of this cell. Parse it,
    #      create the AST, raise FormulaParseError if
    #      there is a problem.  Will be used
    #      when cell this cell is modified by the user
    def setFormula(self, formulaStr):
        if not type(formulaStr)==types.StringType:
            raise TypeError,"setFormula was expecting a string"

        #parsing the formula string, if necessary
        if formulaStr[0]=='=':
            astTuple=getAST(formulaStr+";")
            if not astTuple[0]:
                raise FormulaParseError()
            self.__ast=astTuple[1]
            #creating the dependence list
            v=DependVisitor()
            self.__ast.accept(v)
            self.__dependsOn=v.dependsOn
        else:
            self.__ast=None
            self.__value=formulaStr
            self.__dependsOn={}

    def getFormula(self):
        return self.__formulaStr

    #INPUT: <None>
    #OUTPUT: <listOfTuples:(<int>,<int>)>
    #Order: O(1)
    #Info: Return a list of tuples <column, row>
    #      of the cells that this object depends on
    #      for evaluation
    def getDepend(self):
        return self.__dependsOn.keys()

    #INPUT: <None>
    #OUTPUT: <Float>
    #Order: O(1)
    #Info: Returns the current evaluated value of this cell
    def getValue(self):
        return self.__value

    #INPUT: <None>
    #OUTPUT: <tuple:<int,int>>
    #Order: O(1)
    #Info: Returns the coordinates as a tuple <column,row>
    def getCoords(self):
        return (self.__col,self.__row)

    #INPUT: <None>
    #OUTPUT: <SSheetData>
    #Order: O(1)
    #Info: Returns the spreadsheet data associated with this cell
    def getSSheet(self):
        return self.__SSheetData

    #INPUT: <None>
    #OUTPUT: <None>
    #Order: O(1)
    #Info: Evaluate this cell, if possible.  (i.e. no cycle)
    def evaluate(self):
        if self.__ast==None:
            return self.__value
        elif not self.__SSheetData.canEvaluate(self):
            self.__value="#CYCLE#"
            return self.__value

        v=EvalVisitor(self.__SSheetData)
        self.__ast.accept(v)
        if not v.valueError:
            self.__value=v.stack[0]
        else:
            self.__value="#VALUE!"
        return self.__value

    def __str__(self):
        return str(self.__value)




